@@ -1,26 +1,43 @@
 import inspect
 from functools import wraps
-from inspect import Signature, _empty  # type: ignore
-from typing import Any, List, Sequence, Set, Type, Tuple, TypeVar, Union
+from inspect import Signature, _ParameterKind, _empty  # type: ignore
+from typing import (
+    Any,
+    Awaitable,
+    Callable,
+    ForwardRef,
+    List,
+    Mapping,
+    Optional,
+    Sequence,
+    Set,
+    Tuple,
+    Type,
+    TypeVar,
+    Union,
+)
 from uuid import UUID
 
 from guardpost.authentication import Identity, User
+from rodi import Services
 
+from blacksheep import Request, Response
 from blacksheep.normalization import copy_special_attributes
+from blacksheep.server.routing import Route
 
 from .bindings import (
-    empty,
     Binder,
     BodyBinder,
     BoundValue,
     ControllerBinder,
     ExactBinder,
     IdentityBinder,
-    RequestBinder,
-    QueryBinder,
     JsonBinder,
+    QueryBinder,
+    RequestBinder,
     RouteBinder,
     ServiceBinder,
+    empty,
     get_binder_by_type,
 )
 
@@ -35,18 +52,18 @@ class UnsupportedSignatureError(NormalizationError):
     def __init__(self, method):
         super().__init__(
             f"Cannot normalize method `{method.__qualname__}` because its "
-            f"signature contains *args or *kwargs parameters. "
+            f"signature contains *args, or *kwargs, or keyword only parameters. "
             f"If you use a decorator, please use `functools.@wraps` "
             f"with your wrapper, to fix this error."
         )
 
 
-class MultipleFromBodyParameters(NormalizationError):
-    def __init__(self, method, first_match, new_match):
+class UnsupportedForwardRefInSignatureError(NormalizationError):
+    def __init__(self, unsupported_type):
         super().__init__(
-            f"Cannot use more than one `FromBody` parameters for the same method "
-            f"({method.__qualname__}). The first match was: {first_match}, "
-            f"a second one {new_match}."
+            f"Cannot normalize method `{unsupported_type}` because its "
+            f"signature contains a forward reference (type annotation as string). "
+            f"Use type annotations to exact types to fix this error. "
         )
 
 
@@ -99,7 +116,9 @@ _types_handled_with_query = {
 }
 
 
-def _check_union(parameter, annotation, method):
+def _check_union(
+    parameter: inspect.Parameter, annotation: Any, method: Callable[..., Any]
+) -> Tuple[bool, Any]:
     """
     Checks if the given annotation is Optional[] - in such case unwraps it
     and returns its value.
@@ -128,7 +147,9 @@ def _check_union(parameter, annotation, method):
 
 
 def _get_parameter_binder_without_annotation(
-    parameter, services, route, method, name: str
+    services: Services,
+    route: Optional[Route],
+    name: str,
 ) -> Binder:
     if route:
         # 1. does route contain a parameter with matching name?
@@ -143,7 +164,7 @@ def _get_parameter_binder_without_annotation(
     return QueryBinder(List[str], name, True)
 
 
-def _is_bound_value_annotation(annotation) -> bool:
+def _is_bound_value_annotation(annotation: Any) -> bool:
     if inspect.isclass(annotation) and issubclass(annotation, BoundValue):
         return True
     return "__origin__" in annotation.__dict__ and issubclass(
@@ -176,7 +197,10 @@ def _get_bound_value_type(bound_type: Type[BoundValue]) -> Type[Any]:
 
 
 def _get_parameter_binder(
-    parameter: inspect.Parameter, services, route, method
+    parameter: inspect.Parameter,
+    services: Services,
+    route: Optional[Route],
+    method: Callable[..., Any],
 ) -> Binder:
     name = parameter.name
 
@@ -189,13 +213,14 @@ def _get_parameter_binder(
     original_annotation = parameter.annotation
 
     if original_annotation is _empty:
-        return _get_parameter_binder_without_annotation(
-            parameter, services, route, method, name
-        )
+        return _get_parameter_binder_without_annotation(services, route, name)
 
     # unwrap the Optional[] annotation, if present:
     is_root_optional, annotation = _check_union(parameter, original_annotation, method)
 
+    if isinstance(annotation, (str, ForwardRef)):
+        raise UnsupportedForwardRefInSignatureError(original_annotation)
+
     # 1. is the type annotation of BoundValue[T] type?
     if _is_bound_value_annotation(annotation):
         binder_type = get_binder_by_type(annotation)
@@ -203,6 +228,9 @@ def _get_parameter_binder(
 
         is_optional, expected_type = _check_union(parameter, expected_type, method)
 
+        if isinstance(expected_type, (str, ForwardRef)):
+            raise UnsupportedForwardRefInSignatureError(expected_type)
+
         parameter_name = annotation.name or name
 
         binder = binder_type(expected_type, parameter_name, False)
@@ -223,18 +251,14 @@ def _get_parameter_binder(
 
         return binder
 
-    # 2A. do services contain a service with matching name?
-    if name in services:
-        return ServiceBinder(name, name, True, services)
-
-    # 2B. do services contain a service with matching type?
-    if annotation in services:
-        return ServiceBinder(annotation, annotation.__class__.__name__, True, services)
-
-    # 3. does route contain a parameter with matching name?
+    # 2. does route contain a parameter with matching name?
     if route and name in route.param_names:
         return RouteBinder(annotation, name, True)
 
+    # 3. do services contain a service with matching type?
+    if annotation in services:
+        return ServiceBinder(annotation, annotation.__class__.__name__, True, services)
+
     # 4. is simple type?
     if annotation in _types_handled_with_query:
         return QueryBinder(annotation, name, True, required=not is_root_optional)
@@ -250,7 +274,10 @@ def _get_parameter_binder(
 
 
 def get_parameter_binder(
-    parameter: inspect.Parameter, services, route, method
+    parameter: inspect.Parameter,
+    services: Services,
+    route: Optional[Route],
+    method: Callable[..., Any],
 ) -> Binder:
     binder = _get_parameter_binder(parameter, services, route, method)
     if parameter.default is _empty:
@@ -260,7 +287,9 @@ def get_parameter_binder(
     return binder
 
 
-def _get_binders_for_function(method, services, route) -> List[Binder]:
+def _get_binders_for_function(
+    method: Callable[..., Any], services: Services, route: Optional[Route]
+) -> List[Binder]:
     signature = Signature.from_callable(method)
     parameters = signature.parameters
     body_binder = None
@@ -282,27 +311,25 @@ def _get_binders_for_function(method, services, route) -> List[Binder]:
     return binders
 
 
-def get_binders(route, services) -> Sequence[Binder]:
+def get_binders(route: Route, services: Services) -> List[Binder]:
     """
     Returns a list of binders to extract parameters
     for a request handler.
     """
-    return _get_binders_for_function(route.handler, services, route)
+    binders = _get_binders_for_function(route.handler, services, route)
+    setattr(route.handler, "binders", binders)
+    return binders
 
 
-def get_binders_for_middleware(method, services):
+def get_binders_for_middleware(
+    method: Callable[..., Any], services: Services
+) -> Sequence[Binder]:
     return _get_binders_for_function(method, services, None)
 
 
-def _copy_name_and_docstring(source_method, wrapper):
-    try:
-        wrapper.__name__ = source_method.__name__
-        wrapper.__doc__ = source_method.__doc__
-    except AttributeError:
-        pass
-
-
-def _get_sync_wrapper_for_controller(binders: Sequence[Binder], method):
+def _get_sync_wrapper_for_controller(
+    binders: Sequence[Binder], method: Callable[..., Any]
+) -> Callable[[Request], Awaitable[Response]]:
     @wraps(method)
     async def handler(request):
         values = []
@@ -321,7 +348,9 @@ def _get_sync_wrapper_for_controller(binders: Sequence[Binder], method):
     return handler
 
 
-def _get_async_wrapper_for_controller(binders: Sequence[Binder], method):
+def _get_async_wrapper_for_controller(
+    binders: Sequence[Binder], method: Callable[..., Any]
+) -> Callable[[Request], Awaitable[Response]]:
     @wraps(method)
     async def handler(request):
         values = []
@@ -340,7 +369,13 @@ def _get_async_wrapper_for_controller(binders: Sequence[Binder], method):
     return handler
 
 
-def get_sync_wrapper(services, route, method, params, params_len):
+def get_sync_wrapper(
+    services: Services,
+    route: Route,
+    method: Callable[..., Any],
+    params: Mapping[str, inspect.Parameter],
+    params_len: int,
+) -> Callable[[Request], Awaitable[Response]]:
     if params_len == 0:
         # the user defined a synchronous request handler with no input
         async def handler(_):
@@ -370,7 +405,13 @@ def get_sync_wrapper(services, route, method, params, params_len):
     return handler
 
 
-def get_async_wrapper(services, route, method, params, params_len):
+def get_async_wrapper(
+    services: Services,
+    route: Route,
+    method: Callable[..., Any],
+    params: Mapping[str, inspect.Parameter],
+    params_len: int,
+) -> Callable[[Request], Awaitable[Response]]:
     if params_len == 0:
         # the user defined a request handler with no input
         async def handler(_):
@@ -397,14 +438,19 @@ def get_async_wrapper(services, route, method, params, params_len):
     return handler
 
 
-def normalize_handler(route, services):
+def normalize_handler(
+    route: Route, services: Services
+) -> Callable[[Request], Awaitable[Response]]:
     method = route.handler
 
     sig = Signature.from_callable(method)
     params = sig.parameters
     params_len = len(params)
 
-    if any(str(param).startswith("*") for param in params.values()):
+    if any(
+        str(param).startswith("*") or param.kind.value == _ParameterKind.KEYWORD_ONLY
+        for param in params.values()
+    ):
         raise UnsupportedSignatureError(method)
 
     if inspect.iscoroutinefunction(method):
@@ -413,13 +459,13 @@ def normalize_handler(route, services):
         normalized = get_sync_wrapper(services, route, method, params, params_len)
 
     if normalized is not method:
+        setattr(normalized, "root_fn", method)
         copy_special_attributes(method, normalized)
-        _copy_name_and_docstring(method, normalized)
 
     return normalized
 
 
-def _is_basic_middleware_signature(parameters):
+def _is_basic_middleware_signature(parameters: Mapping[str, inspect.Parameter]) -> bool:
     values = list(parameters.values())
 
     if len(values) != 2:
@@ -432,7 +478,9 @@ def _is_basic_middleware_signature(parameters):
     return False
 
 
-def _get_middleware_async_binder(method, services):
+def _get_middleware_async_binder(
+    method: Callable[..., Awaitable[Response]], services: Services
+) -> Callable[[Request, Callable[..., Any]], Awaitable[Response]]:
     binders = get_binders_for_middleware(method, services)
 
     async def handler(request, next_handler):
@@ -456,16 +504,16 @@ def _get_middleware_async_binder(method, services):
     return handler
 
 
-def normalize_middleware(middleware, services):
+def normalize_middleware(
+    middleware: Callable[..., Awaitable[Response]], services: Services
+) -> Callable[[Request, Callable[..., Any]], Awaitable[Response]]:
+    if not inspect.iscoroutinefunction(middleware):
+        raise ValueError("Middlewares must be asynchronous functions")
+
     sig = Signature.from_callable(middleware)
     params = sig.parameters
 
     if _is_basic_middleware_signature(params):
         return middleware
 
-    if inspect.iscoroutinefunction(middleware):
-        normalized = _get_middleware_async_binder(middleware, services)
-    else:
-        raise ValueError("Middlewares must be asynchronous functions")
-
-    return normalized
+    return _get_middleware_async_binder(middleware, services)
