@@ -1,4 +1,5 @@
 import re
+import logging
 from abc import abstractmethod
 from collections import defaultdict
 from functools import lru_cache
@@ -6,7 +7,7 @@ from typing import Any, Callable, Dict, List, Optional, AnyStr
 from urllib.parse import unquote
 
 from blacksheep import HttpMethod
-from blacksheep.utils import BytesOrStr, ensure_bytes, ensure_str
+from blacksheep.utils import ensure_bytes, ensure_str
 
 __all__ = [
     "Router",
@@ -20,6 +21,7 @@ __all__ = [
 
 _route_all_rx = re.compile(b"\\*")
 _route_param_rx = re.compile(b"/:([^/]+)")
+_mustache_route_param_rx = re.compile(b"/{([^}]+)}")
 _named_group_rx = re.compile(b"\\?P<([^>]+)>")
 _escaped_chars = {b".", b"[", b"]", b"(", b")"}
 
@@ -30,7 +32,17 @@ def _get_regex_for_pattern(pattern):
         if c in pattern:
             pattern = pattern.replace(c, b"\\" + c)
     if b"*" in pattern:
-        pattern = _route_all_rx.sub(br"(?P<tail>.+)", pattern)
+        # throw exception if a star appears more than once
+        if pattern.count(b"*") > 1:
+            raise ValueError(
+                "A route pattern cannot contain more than one star sign *. "
+                "Multiple star signs are not supported."
+            )
+
+        if b"/*" in pattern:
+            pattern = _route_all_rx.sub(br"?(?P<tail>.*)", pattern)
+        else:
+            pattern = _route_all_rx.sub(br"(?P<tail>.*)", pattern)
     if b"/:" in pattern:
         pattern = _route_param_rx.sub(br"/(?P<\1>[^\/]+)", pattern)
 
@@ -86,34 +98,51 @@ class RouteMatch:
             else None
         )
 
-    def __repr__(self):
-        return f"<RouteMatch {id(self)}>"
-
 
 class Route:
 
     __slots__ = ("handler", "pattern", "has_params", "param_names", "_rx")
 
+    pattern: bytes
+
     def __init__(self, pattern: AnyStr, handler: Any):
-        if isinstance(pattern, str):
-            pattern = pattern.encode("utf8")
-        if pattern == b"":
-            pattern = b"/"
-        if len(pattern) > 1 and pattern.endswith(b"/"):
-            pattern = pattern.rstrip(b"/")
-        pattern = pattern.lower()
+        raw_pattern = self.normalize_pattern(pattern)
         self.handler = handler
-        self.pattern = pattern
-        self.has_params = b"*" in pattern or b":" in pattern
-        rx, param_names = _get_regex_for_pattern(pattern)
+        self.pattern = raw_pattern
+        self.has_params = b"*" in raw_pattern or b":" in raw_pattern
+        rx, param_names = _get_regex_for_pattern(raw_pattern)
         self._rx = rx
         self.param_names = [name.decode("utf8") for name in param_names]
 
+    def normalize_pattern(self, pattern: AnyStr) -> bytes:
+        if isinstance(pattern, str):
+            raw_pattern = pattern.encode("utf8")
+        else:
+            raw_pattern = pattern
+
+        # support for mustache patterns, e.g. /api/cats/{cat_id}
+        if b"{" in raw_pattern:
+            raw_pattern = _mustache_route_param_rx.sub(br"/:\1", raw_pattern)
+
+        if raw_pattern == b"":
+            raw_pattern = b"/"
+        if len(raw_pattern) > 1 and raw_pattern.endswith(b"/"):
+            raw_pattern = raw_pattern.rstrip(b"/")
+
+        return raw_pattern.lower()
+
+    def __repr__(self) -> str:
+        return f"<Route {self.pattern.decode('utf8')}>"
+
+    @property
+    def mustache_pattern(self) -> str:
+        return _route_param_rx.sub(br"/{\1}", self.pattern).decode("utf8")
+
     @property
     def full_pattern(self) -> bytes:
         return self._rx.pattern
 
-    def match(self, value: bytes):
+    def match(self, value: bytes) -> Optional[RouteMatch]:
         if not self.has_params and value.lower() == self.pattern:
             return RouteMatch(self, None)
 
@@ -124,22 +153,21 @@ class Route:
 
         return RouteMatch(self, match.groupdict() if self.has_params else None)
 
-    def __repr__(self):
-        return f"<Route {self.pattern}>"
-
 
 class RouterBase:
     @abstractmethod
-    def add(self, method: str, pattern: BytesOrStr, handler: Callable):
-        ...
+    def add(self, method: str, pattern: str, handler: Callable) -> None:
+        """Adds a request handler for the given HTTP method and route pattern."""
 
-    def mark_handler(self, handler: Callable):
+    def mark_handler(self, handler: Callable) -> None:
         setattr(handler, "route_handler", True)
 
-    def normalize_default_pattern_name(self, handler_name: str):
+    def normalize_default_pattern_name(self, handler_name: str) -> str:
         return handler_name.replace("_", "-")
 
-    def get_decorator(self, method, pattern="/"):
+    def get_decorator(
+        self, method: str, pattern: Optional[str] = "/"
+    ) -> Callable[..., Any]:
         def decorator(fn):
             nonlocal pattern
             if pattern is ... or pattern is None:
@@ -149,69 +177,68 @@ class RouterBase:
                 else:
                     pattern = "/" + self.normalize_default_pattern_name(fn.__name__)
 
-                # TODO: implement log here
-                # app_logger.info('Defaulting to route pattern "%s" for
-                # request handler <%s>', pattern, fn.__qualname__)
+                logging.debug(
+                    "Defaulting to route pattern '%s' for" "request handler <%s>",
+                    pattern,
+                    fn.__qualname__,
+                )
             self.add(method, pattern, fn)
             return fn
 
         return decorator
 
-    def add_head(self, pattern, handler):
+    def add_head(self, pattern: str, handler: Callable[..., Any]) -> None:
         self.add(HttpMethod.HEAD, pattern, handler)
 
-    def add_get(self, pattern, handler):
+    def add_get(self, pattern: str, handler: Callable[..., Any]) -> None:
         self.add(HttpMethod.GET, pattern, handler)
 
-    def add_post(self, pattern, handler):
+    def add_post(self, pattern: str, handler: Callable[..., Any]) -> None:
         self.add(HttpMethod.POST, pattern, handler)
 
-    def add_put(self, pattern, handler):
+    def add_put(self, pattern: str, handler: Callable[..., Any]) -> None:
         self.add(HttpMethod.PUT, pattern, handler)
 
-    def add_delete(self, pattern, handler):
+    def add_delete(self, pattern: str, handler: Callable[..., Any]) -> None:
         self.add(HttpMethod.DELETE, pattern, handler)
 
-    def add_trace(self, pattern, handler):
+    def add_trace(self, pattern: str, handler: Callable[..., Any]) -> None:
         self.add(HttpMethod.TRACE, pattern, handler)
 
-    def add_options(self, pattern, handler):
+    def add_options(self, pattern: str, handler: Callable[..., Any]) -> None:
         self.add(HttpMethod.OPTIONS, pattern, handler)
 
-    def add_connect(self, pattern, handler):
+    def add_connect(self, pattern: str, handler: Callable[..., Any]) -> None:
         self.add(HttpMethod.CONNECT, pattern, handler)
 
-    def add_patch(self, pattern, handler):
+    def add_patch(self, pattern: str, handler: Callable[..., Any]) -> None:
         self.add(HttpMethod.PATCH, pattern, handler)
 
-    def add_any(self, pattern, handler):
-        self.add("*", pattern, handler)
-
-    def head(self, pattern="/"):
+    def head(self, pattern: Optional[str] = "/") -> Callable[..., Any]:
         return self.get_decorator(HttpMethod.HEAD, pattern)
 
-    def get(self, pattern="/"):
+    def get(self, pattern: Optional[str] = "/") -> Callable[..., Any]:
         return self.get_decorator(HttpMethod.GET, pattern)
 
-    def post(self, pattern="/"):
+    def post(self, pattern: Optional[str] = "/") -> Callable[..., Any]:
         return self.get_decorator(HttpMethod.POST, pattern)
 
-    def put(self, pattern="/"):
+    def put(self, pattern: Optional[str] = "/") -> Callable[..., Any]:
         return self.get_decorator(HttpMethod.PUT, pattern)
 
-    def delete(self, pattern="/"):
+    def delete(self, pattern: Optional[str] = "/") -> Callable[..., Any]:
         return self.get_decorator(HttpMethod.DELETE, pattern)
 
-    def trace(self, pattern="/"):
+    def trace(self, pattern: Optional[str] = "/") -> Callable[..., Any]:
         return self.get_decorator(HttpMethod.TRACE, pattern)
 
-    def options(self, pattern="/"):
+    def options(self, pattern: Optional[str] = "/") -> Callable[..., Any]:
         return self.get_decorator(HttpMethod.OPTIONS, pattern)
 
-    def connect(self, pattern="/"):
+    def connect(self, pattern: Optional[str] = "/") -> Callable[..., Any]:
         return self.get_decorator(HttpMethod.CONNECT, pattern)
 
-    def patch(self, pattern="/"):
+    def patch(self, pattern: Optional[str] = "/") -> Callable[..., Any]:
         return self.get_decorator(HttpMethod.PATCH, pattern)
 
 
@@ -234,7 +261,7 @@ class Router(RouterBase):
             if callable(value):
                 self._fallback = Route(b"*", value)
                 return
-            raise ValueError("fallback must be a Route")
+            raise ValueError("fallback must be a Route or a callable")
         self._fallback = value
 
     def __iter__(self):
@@ -265,15 +292,14 @@ class Router(RouterBase):
             raise RouteDuplicate(method, new_route.pattern, current_route.handler)
         self._set_configured_route(method, new_route)
 
-    def add(self, method: str, pattern: BytesOrStr, handler: Any):
+    def add(self, method: str, pattern: AnyStr, handler: Any):
         self.mark_handler(handler)
         method_name = ensure_bytes(method)
-        pattern = ensure_bytes(pattern)
-        new_route = Route(pattern, handler)
+        new_route = Route(ensure_bytes(pattern), handler)
         self._check_duplicate(method_name, new_route)
         self.add_route(method_name, new_route)
 
-    def add_route(self, method: BytesOrStr, route: Route):
+    def add_route(self, method: AnyStr, route: Route):
         self.routes[ensure_bytes(method)].append(route)
 
     def sort_routes(self):
@@ -285,18 +311,15 @@ class Router(RouterBase):
 
         for method in current_routes.keys():
             current_routes[method].sort(key=lambda route: len(route.param_names))
-
+            current_routes[method].sort(key=lambda route: b"*" in route.pattern)
             current_routes[method].sort(key=lambda route: b"*" == route.pattern)
 
         self.routes = current_routes
 
     @lru_cache(maxsize=1200)
-    def get_match(self, method: BytesOrStr, value: BytesOrStr) -> Optional[RouteMatch]:
-        method = ensure_bytes(method)
-        value = ensure_bytes(value)
-
-        for route in self.routes[method]:
-            match = route.match(value)
+    def get_match(self, method: AnyStr, value: AnyStr) -> Optional[RouteMatch]:
+        for route in self.routes[ensure_bytes(method)]:
+            match = route.match(ensure_bytes(value))
             if match:
                 return match
         if self._fallback is None:
@@ -309,17 +332,11 @@ class RegisteredRoute:
 
     __slots__ = ("method", "pattern", "handler")
 
-    def __init__(self, method: str, pattern: BytesOrStr, handler: Callable):
+    def __init__(self, method: str, pattern: str, handler: Callable):
         self.method = method
         self.pattern = pattern
         self.handler = handler
 
-    def __repr__(self):
-        return (
-            f'<RegisteredRoute {self.method} "{self.pattern}" '
-            f"{self.handler.__name__}>"
-        )
-
 
 class RoutesRegistry(RouterBase):
     """
@@ -336,9 +353,6 @@ class RoutesRegistry(RouterBase):
     def __iter__(self):
         yield from self.routes
 
-    def add(self, method: str, pattern: BytesOrStr, handler: Callable):
+    def add(self, method: str, pattern: str, handler: Callable):
         self.mark_handler(handler)
         self.routes.append(RegisteredRoute(method, pattern, handler))
-
-    def __repr__(self):
-        return f"<RoutesRegistry {self.routes}>"
